use crate::grid::{Grid, Pos};

// This algorithm is easier to implement when done in octants, and the steps are exactly the same for each octant.
// This struct translates coordinates so that the code for one octant can be used on all of them.
struct Octant<'a, T> {
    grid: &'a mut T,
    origin: Pos,
    transform: (isize, isize, isize, isize),
}

impl<'a, T> Octant<'a, T> {
    fn new(grid: &'a mut T, origin: Pos) -> Self {
        Self {
            grid,
            origin,
            transform: (1, 0, 0, 1),
        }
    }

    /// Rotate the internal transform by 90 degrees.
    fn rot(&mut self) {
        let (a, b, c, d) = self.transform;
        self.transform = (-c, -d, a, b);
    }

    /// Flip the internal transform horizontally.
    fn flip(&mut self) {
        let (a, b, c, d) = self.transform;
        self.transform = (-a, -b, c, d);
    }

    fn transform(&self, x: isize, y: isize) -> Pos {
        let (a, b, c, d) = self.transform;

        (self.origin.0 + a * x + b * y, self.origin.1 + c * x + d * y)
    }
}

impl<T: Grid> Octant<'_, T> {
    fn is_wall(&self, x: isize, y: isize) -> bool {
        self.grid.is_wall(self.transform(x, y))
    }

    fn mark_visible(&mut self, x: isize, y: isize) {
        self.grid.mark_visible(self.transform(x, y))
    }
}

// x ------>
//
// -------------- <- top (slope)
// \
//  \                       y
//   \                      |
//    \                     |
//     \                    |
//      \                   v
//       \
//        \
//         \
//          \
//           \
//            \
//             \
//              \ <- bottom (slope)

// upper_scope and lower_slope find the upper and lower bounds of slope required to "hit"
// a given square, taking into account bevelling.

//       possible upper slopes
//           /---------\
//           v         v
// /---------^---------\
// |        / \        |
// |       /   \       |
// |      /     \      |
// |     /       \     |
// |    /         \    |
// |   /           \   |
// |  /             \  |
// | /               \ |
// |/                 \|
// |\                 /|
// | \               / |
// |  \             /  |
// |   \           /   |
// |    \         /    |
// |     \       /     |
// |      \     /      |
// |       \   /       |
// |        \ /        |
// \---------v---------/
fn upper_slope<T: Grid>(oct: &Octant<T>, x: isize, y: isize) -> f64 {
    let mut a = x as f64;
    let b = y as f64 - 0.5;
    if oct.is_wall(x + 1, y) || oct.is_wall(x, y - 1) || oct.is_wall(x + 1, y - 1) {
        a += 0.5;
    }
    b / a
}

//     /---------^---------\
//     |        / \        |
//     |       /   \       |
//     |      /     \      |
//     |     /       \     |
//     |    /         \    |
//     |   /           \   |
//     |  /             \  |
//     | /               \ |
//     |/                 \|
//     |\                 /|
//     | \               / |
//     |  \             /  |
//     |   \           /   |
//     |    \         /    |
//     |     \       /     |
//     |      \     /      |
//     |       \   /       |
//     |        \ /        |
//     \---------v---------/
//     ^         ^
//     \---------/
// possible lower slopes
fn lower_slope<T: Grid>(oct: &Octant<T>, x: isize, y: isize) -> f64 {
    let mut a = x as f64;
    let b = y as f64 + 0.5;
    if oct.is_wall(x - 1, y) || oct.is_wall(x, y + 1) || oct.is_wall(x - 1, y + 1) {
        a -= 0.5;
    }
    b / a
}

fn hitting<T: Grid>(oct: &Octant<T>, x: isize, y: isize, slope: f64) -> bool {
    oct.is_wall(x, y) && upper_slope(oct, x, y) < slope && slope < lower_slope(oct, x, y)
}

fn cast<T: Grid>(oct: &mut Octant<T>, x: isize, top: f64, bottom: f64) {
    let mut start = (top * x as f64).ceil() as isize;
    let mut end = (bottom * x as f64).floor() as isize;

    if hitting(oct, x, start - 1, top) {
        start -= 1;
    }
    if hitting(oct, x, end + 1, bottom) {
        end += 1;
    }

    let mut next_top = top;
    let mut last_was_wall = oct.is_wall(x, start);
    oct.mark_visible(x, start);

    for y in start + 1..=end {
        let this_is_wall = oct.is_wall(x, y);

        if last_was_wall && !this_is_wall {
            next_top = lower_slope(oct, x, y - 1);
        } else if !last_was_wall && this_is_wall {
            cast(oct, x + 1, next_top, upper_slope(oct, x, y));
        }

        if next_top == bottom {
            break;
        }
        oct.mark_visible(x, y);

        last_was_wall = this_is_wall;
    }

    if !last_was_wall {
        cast(oct, x + 1, next_top, bottom);
    }
}

/// Mark the visible points on a grid.
///
/// Calls [`Grid::mark_visible`] for each cell in `grid` visible from `origin`, including walls, non-walls, and the origin itself.
pub fn mark_vision<T: Grid>(grid: &mut T, origin: Pos) {
    grid.mark_visible(origin);
    let mut oct = Octant::new(grid, origin);
    for _ in 0..2 {
        for _ in 0..4 {
            cast(&mut oct, 1, 0.0, 1.0);
            oct.rot();
        }
        oct.flip();
    }
}
